MODULE draft;

    (* For trying out miscellaneous ideas. *)


FROM SYSTEM IMPORT CAST, CARD32;

FROM LONGLONG IMPORT
    (* const*)  Zero64,
    (* type *)  CARD64, INT64,
    (* proc *)  Mul32, Div10, LongDiv64;

FROM LONG128 IMPORT
    (* type *)  CARD128,
    (* proc *)  IsZero, LongDiv128;

FROM RandCard IMPORT
    (* proc *)  Randomize, RandCardinal;

FROM STextIO IMPORT
    (* proc *)  WriteChar, WriteString, WriteLn;

(************************************************************************)
(*                           OUTPUT FUNCTIONS                           *)
(************************************************************************)

PROCEDURE WriteCard (N: CARDINAL);

    (* Write N in decimal. *)

    BEGIN
        IF N > 9 THEN
            WriteCard (N DIV 10);  N := N MOD 10;
        END (*IF*);
        WriteChar (CHR(N + ORD('0')));
    END WriteCard;

(************************************************************************)

PROCEDURE WriteInt (N: INTEGER);

    (* Write N in decimal. *)

    BEGIN
        IF N < 0 THEN
            WriteChar ('-');
            N := -N;
        END (*IF*);
        WriteCard (N);
    END WriteInt;

(************************************************************************)

PROCEDURE WriteCard64 (N: CARD64);

    (* Write N in decimal. *)

    VAR Q: CARD64;  R: CARDINAL;

    BEGIN
        IF (N.high <> 0) OR (N.low > 9) THEN
            Div10 (N, Q, R);
            WriteCard64 (Q);  N := CARD64 {0,R};
        END (*IF*);
        WriteChar (CHR(N.low + ORD('0')));
    END WriteCard64;

(************************************************************************)

PROCEDURE WriteHex1 (N: CARD32);

    (* Write N as a 1-digit hexadecimal number. *)

    BEGIN
        IF N > 9 THEN
            WriteChar (CHR(N - 10 + ORD('A')));
        ELSE
            WriteChar (CHR(N + ORD('0')));
        END (*IF*);
    END WriteHex1;

(************************************************************************)

PROCEDURE WriteHexk (N, k: CARD32);

    (* Write N as a k-digit hexadecimal number. *)

    BEGIN
        IF k > 1 THEN
            WriteHexk (N DIV 16, k-1);
            N := N MOD 16;
        END (*IF*);
        WriteHex1 (N);
    END WriteHexk;

(************************************************************************)

PROCEDURE WriteHex32 (N: CARD32);

    (* Write N in hexadecimal. *)

    BEGIN
        WriteHexk (N, 8);
    END WriteHex32;

(************************************************************************)

PROCEDURE WriteHex64 (N: CARD64);

    (* Write N in hexadecimal. *)

    BEGIN
        WriteHex32 (N.high);
        WriteHex32 (N.low);
        WriteChar ('H');
    END WriteHex64;

(************************************************************************)

PROCEDURE WriteHex32LJ (N: CARD32);

    (* Write N in hexadecimal, left justified. *)

    VAR Q: CARD64;  R: CARDINAL;

    BEGIN
        IF N > 15 THEN
            WriteHex32LJ (N DIV 16);  N := N MOD 16;
        END (*IF*);
        WriteHex1 (N);
    END WriteHex32LJ;

(************************************************************************)

PROCEDURE WriteHex64LJ (N: CARD64);

    (* Write N in hexadecimal, left justified. *)

    VAR Q: CARD64;  R: CARDINAL;

    BEGIN
        IF (N.high <> 0) OR (N.low > 15) THEN
            LongDiv64 (N, 16, Q, R);
            WriteHex64LJ (Q);  N := CARD64 {0,R};
        END (*IF*);
        WriteHex1 (N.low);
    END WriteHex64LJ;

(************************************************************************)
(*                           MODULAR ARITHMETIC                         *)
(************************************************************************)

PROCEDURE gcd (a, b: CARDINAL): CARDINAL;

    (* Greatest common divisor. *)

    VAR temp: CARDINAL;

    BEGIN
        IF a = b THEN RETURN a;
        ELSIF b > a THEN
            temp := b;  b := a;  a := temp;
        END (*IF*);
        (* Now a > b for the rest of the calculation. *)
        WHILE b > 0 DO
            temp := a;
            a := b;
            b := temp MOD b;
        END (*WHILE*);
        RETURN a;
    END gcd;

(************************************************************************)

PROCEDURE xgcd (VAR (*OUT*) result: ARRAY OF INTEGER;  a, b: CARDINAL);

    (* Extended gcd computation.  The result is a triple (g, x, y)      *)
    (* such that ax + by = g, and g = gcd(a,b).                         *)

    VAR q: INTEGER;
        s0, s1, t0, t1: INTEGER;

    BEGIN
        s0 := 1;  s1 := 0;
        t0 := 0;  t1 := 1;
        LOOP
            q := a DIV b;  a := a MOD b;
            DEC (s0, q*s1);
            DEC (t0, q*t1);
            IF a = 0 THEN
                result[0] := b; result[1] := s1; result[2] := t1;
                RETURN;
            END (*IF*);
            q := b DIV a;  b := b MOD a;
            DEC (s1, q*s0);
            DEC (t1, q*t0);
            IF b = 0 THEN
                result[0] := a; result[1] := s0; result[2] := t0;
                RETURN;
            END (*IF*);
        END (*LOOP*);
    END xgcd;

(************************************************************************)

PROCEDURE inverse (a: CARDINAL;  m: CARD64): CARD64;

    (* Returns x such that ax = 1 mod m.  This will work only if a and  *)
    (* m are relatively prime.                                          *)

    (* The extended Euclid algorithm computes a triple (g, x, y) such   *)
    (* that ax + my = g, and g = gcd(a,m).  This modified version       *)
    (* returns only x, since all we want is a solution of ax = g mod m. *)

    VAR t0, t1, q: INTEGER;
        s0, s1: INT64;
        x, originalm: CARD64;

    BEGIN
        (*
        originalm := m;
        s0 := INT64{0,1};  s1 := INT64{0,0};
        t0 := 0;  t1 := 1;
        LOOP
            q := a DIV m;  a := a MOD m;
            DEC (s0, q*s1);
            DEC (t0, q*t1);
            IF a = 0 THEN
                x := s1;
                EXIT (*LOOP*);
            END (*IF*);
            q := m DIV a;  m := m MOD a;
            DEC (s1, q*s0);
            DEC (t1, q*t0);
            IF m = 0 THEN
                x := s0;
                EXIT (*LOOP*);
            END (*IF*);
        END (*LOOP*);
        IF x < 0 THEN INC (x, originalm);
        END (*IF*);
        *)
        x.high := 0;
        x.low := 1;
        RETURN x;
    END inverse;

(************************************************************************)

PROCEDURE VLongDiv64 (A, B: CARD64;  VAR (*OUT*) quotient: CARDINAL;  remainder: CARD64);

    (* Divides A by B, returns quotient and remainder. *)

    CONST mask = MAX(CARDINAL) DIV 2 + 1;    (* 2^31 *)

    VAR AA, RR: CARD128;  Q, QQ, R, BZ: CARD64;  Z: CARDINAL;
        carry: BOOLEAN;

    BEGIN
        AA := CARD128 {Zero64, A};
        LongDiv128 (AA, B, QQ, RR);
        quotient := QQ.low;
        IF QQ.high <> 0 THEN
            WriteString (" Overflow in VLongDiv64");  WriteLn;
        END (*IF*);
        remainder := RR.low;
        IF NOT IsZero(RR) THEN
            WriteString (" Overflow in VLongDiv64");  WriteLn;
        END (*IF*);
    END VLongDiv64;

(************************************************************************)

PROCEDURE EEA (a, b: CARDINAL;  VAR (*OUT*) x, y: INTEGER);

    (* Extended Euclid algorithm, an implementation from first principles. *)
    (* We calculate x and y such that ax+by = gcd(a,b).                    *)

    VAR oldR, R, newR: CARDINAL;
        xold, xnew: INTEGER;
        yold, ynew: INTEGER;
        aint, bint: INTEGER;
        Q: CARDINAL;

    BEGIN
        aint := a;  bint := b;
        oldR := a;  R := b;
        xold := 1;  x := 0;
        yold := 0;  y := 1;
        LOOP
            (*
            WriteString ("R, x, y, a*x+b*y = ");  WriteInt(R);
            WriteString (", ");  WriteInt(x);
            WriteString (", ");  WriteInt(y);
            WriteString (", ");  WriteInt(aint*x+bint*y);
            WriteLn;
            *)
            Q := oldR DIV R;  newR := oldR MOD R;
            xnew := xold - VAL(INTEGER,Q)*x;
            ynew := yold - VAL(INTEGER,Q)*y;
            IF newR = 0 THEN EXIT(*LOOP*) END(*LOOP*);
            oldR := R;  R := newR;
            xold := x;  x := xnew;
            yold := y;  y := ynew;
        END (*LOOP*);
    END EEA;

(************************************************************************)

PROCEDURE EEA1 (a, m: CARDINAL): CARDINAL;

    (* A variant of EEA where we don't care about the solution for y.   *)
    (* This is of interest mostly in the case where a and m are         *)
    (* relatively prime, because in that case x = inverse(a) mod m.     *)

    VAR oldR, R, newR: CARDINAL;
        xold, x, xnew: INTEGER;
        Q: CARDINAL;

        (* The "old" variables are here because this is a second-order  *)
        (* algorithm where we have to look two steps backwards.         *)

    BEGIN
        oldR := a;  R := m;
        xold := 1;  x := 0;
        LOOP
            Q := oldR DIV R;  newR := oldR MOD R;
            xnew := xold - VAL(INTEGER,Q)*x;
            IF newR = 0 THEN EXIT(*LOOP*) END(*LOOP*);
            oldR := R;  R := newR;
            xold := x;  x := xnew;
        END (*LOOP*);
        IF x < 0 THEN
            RETURN x+VAL(INTEGER, m);
        ELSE
            RETURN x;
        END (*IF*);
    END EEA1;

(************************************************************************)

PROCEDURE EEA1a (a, b: CARDINAL;  VAR (*OUT*) y: INTEGER);

    (* Extended Euclid algorithm, an implementation from first principles. *)
    (* We calculate x and y such that ax+by = gcd(a,b).                    *)

    (* A variant of EEA where we don't care about the solution for x.   *)
    (* This is of interest mostly in the case where a and b are         *)
    (* relatively prime, because in that case y = inverse(b) mod a.     *)

    VAR oldR, R, newR: CARDINAL;
        yold, ynew: INTEGER;
        aint, bint: INTEGER;
        Q: CARDINAL;

    BEGIN
        aint := a;  bint := b;
        oldR := a;  R := b;
        yold := 0;  y := 1;
        LOOP
            (*
            WriteString ("R, x, y, a*x+b*y = ");  WriteInt(R);
            WriteString (", ");  WriteInt(x);
            WriteString (", ");  WriteInt(y);
            WriteString (", ");  WriteInt(aint*x+bint*y);
            WriteLn;
            *)
            Q := oldR DIV R;  newR := oldR MOD R;
            ynew := yold - VAL(INTEGER,Q)*y;
            IF newR = 0 THEN EXIT(*LOOP*) END(*LOOP*);
            oldR := R;  R := newR;
            yold := y;  y := ynew;
        END (*LOOP*);
    END EEA1a;

(************************************************************************)

PROCEDURE EEA1b (a: CARD64;  b: CARDINAL;  VAR (*OUT*) y: INTEGER);

    (* Extended Euclid algorithm, an implementation from first principles. *)
    (* We calculate x and y such that ax+by = gcd(a,b).                    *)

    (* A variant of EEA where we don't care about the solution for x.   *)
    (* This is of interest mostly in the case where a and b are         *)
    (* relatively prime, because in that case y = inverse(b) mod a.     *)

    VAR oldR, R, newR: CARD64;
        yold, ynew: INTEGER;
        aint: INT64;  bint: INTEGER;
        Q: CARDINAL;

    BEGIN
        aint := CAST(INT64,a);  bint := b;
        oldR := a;  R := CARD64{0,b};
        yold := 0;  y := 1;
        LOOP
            (*
            WriteString ("R, x, y, a*x+b*y = ");  WriteInt(R);
            WriteString (", ");  WriteInt(x);
            WriteString (", ");  WriteInt(y);
            WriteString (", ");  WriteInt(aint*x+bint*y);
            WriteLn;
            *)
            VLongDiv64 (oldR, R, Q, R);
            (*Q := oldR DIV R;  newR := oldR MOD R;*)
            ynew := yold - VAL(INTEGER,Q)*y;
            IF (newR.high = Zero64.high) AND (newR.low = Zero64.low) THEN
                EXIT(*LOOP*)
            END(*LOOP*);
            oldR := R;  R := newR;
            yold := y;  y := ynew;
        END (*LOOP*);
    END EEA1b;

(************************************************************************)

(*
PROCEDURE EEA2 (a:CARD64;  m: CARD64): CARDINAL;

    (* A variant of EEA where we don't care about the solution for y.   *)
    (* This is of interest mostly in the case where a and m are         *)
    (* relatively prime, because in that case x = inverse(a) mod m.     *)

    VAR oldR, R, newR: CARD64;
        xold, x, xnew: INTEGER;
        Q: CARDINAL;

        (* The "old" variables are here because this is a second-order  *)
        (* algorithm where we have to look two steps backwards.         *)

    BEGIN
        oldR := a;  R := m;
        xold := 1;  x := 0;
        LOOP
            Q := oldR DIV R;  newR := oldR MOD R;
            xnew := xold - VAL(INTEGER,Q)*x;
            IF newR = 0 THEN EXIT(*LOOP*) END(*LOOP*);
            oldR := R;  R := newR;
            xold := x;  x := xnew;
        END (*LOOP*);
        IF x < 0 THEN
            RETURN x+VAL(INTEGER, m);
        ELSE
            RETURN x;
        END (*IF*);
    END EEA2;
*)

(************************************************************************)

PROCEDURE lcm (a, b: CARDINAL): CARD64;

    (* Least common multiple. *)

    BEGIN
        IF a = b THEN RETURN CARD64{0,a};
        ELSE RETURN Mul32(a DIV gcd(a,b), b);
        END (*IF*);
    END lcm;

(************************************************************************)

PROCEDURE ModularPower (base, exponent, modulus: CARDINAL): CARDINAL;

    (* Calculates base^exponent MOD modulus.  We use the traditional    *)
    (* fast squaring method, but with modular reduction to stop the     *)
    (* intermediate results from getting too large.                     *)

    VAR result: CARDINAL;  product, Q: CARD64;

    BEGIN
        IF modulus = 1 THEN RETURN 0 END(*IF*);
        result := 1;
        base := base MOD modulus;
        WHILE exponent > 0 DO
            IF ODD (exponent) THEN
                product := Mul32 (result, base);
                LongDiv64 (product, modulus, Q, result);
            END (*IF*);
            exponent := exponent DIV 2;
            product := Mul32 (base, base);
            LongDiv64 (product, modulus, Q, base);
        END (*WHILE*);
        RETURN result;
    END ModularPower;

(************************************************************************)
(*                            PRIME NUMBERS                             *)
(************************************************************************)

PROCEDURE ProbablePrime (N: CARDINAL): BOOLEAN;

    (* Returns TRUE if there is a high probability that N is prime.     *)
    (* We use the SPRP (strong probable prime) condition, using base 2. *)

    CONST A = 2;

    VAR s, pwr, d, r: CARDINAL;

    BEGIN
        IF N <= 1 THEN RETURN FALSE END(*IF*);

        (* Find an s such that N < 2^s.  That will be the starting  *)
        (* point for the next step.                                 *)

        s := 1;  pwr := 2;
        WHILE pwr <= N DIV 2 DO
            INC (s);  pwr := 2*pwr;
        END (*WHILE*);

        (* Find d and s for which N-1 = d*2^s.   The loop below     *)
        (* will certainly terminate, with s = 0 in the worst case.  *)

        d := (N-1) DIV pwr;
        WHILE d*pwr <> N-1 DO
            DEC (s);  pwr := pwr DIV 2;
            d := (N-1) DIV pwr;
        END (*WHILE*);

        (* Check whether A^d = 1 MOD N.  *)

        pwr := ModularPower (A, d, N);
        IF pwr = 1 THEN RETURN TRUE END(*IF*);

        (* Check whether A^(d*2^r) = N-1 MOD N for some r in [0..s-1].  *)
        (* For this calculation pwr = 2^r.                              *)

        r := 0;  pwr := 1;
        LOOP
            IF r = s THEN RETURN FALSE END(*IF*);
            IF ModularPower (A, d*pwr, N) = N-1 THEN RETURN TRUE END(*IF*);
            INC (r);  pwr := 2*pwr;
        END (*LOOP*);

    END ProbablePrime;

(************************************************************************)

PROCEDURE RandomPrime(): CARDINAL;

    (* Returns a random probably prime between max and min.  This is    *)
    (* just for proof of concept.  I'll probably have to come up with   *)
    (* a better random number generator.                                *)

    CONST
        min = MAX(CARDINAL) DIV 4;  max = MAX(CARDINAL);
        scale = FLOAT(max-min) / FLOAT(MAX(CARDINAL));

    VAR p: CARDINAL;

    BEGIN
        REPEAT
            p := RandCardinal();

            (* Scale to desired interval. *)

            p := VAL(CARDINAL,scale*FLOAT(p)) + min;
        UNTIL ProbablePrime (p);
        RETURN p;
    END RandomPrime;

(************************************************************************)
(*                             RSA EXAMPLE                              *)
(************************************************************************)

PROCEDURE RSA_Generate_example;

    (* This uses artificially small numbers, to let me see what is      *)
    (* going on without having to deal with the complexities of very    *)
    (* wide numbers.                                                    *)

    VAR p, q, e, R: CARDINAL;  lambda, n, Q: CARD64;
        d: INTEGER;

    BEGIN
        WriteLn;  WriteString ("RSA example:");  WriteLn;

        (* Choose two distinct prime numbers p and q.  Let n = p*q *)

        p := RandomPrime();  q := RandomPrime();
        WriteString ("We have chosen two random primes, p = ");
        WriteCard (p);  WriteString (" and q = ");  WriteCard (q);
        WriteLn;
        n := Mul32 (p, q);

        lambda := lcm (p-1, q-1);

        (* Choose e such that 1 < e < lambda that is coprime to lambda. *)
        (* Making e prime simplifies the coprime check.                 *)

        e := 17;
        LongDiv64 (lambda, e, Q, R);
        IF R = 0 THEN
            WriteString ("Poorly chosen e. Try again.");  WriteLn;
        END (*IF*);

        WriteString ("lambda = ");  WriteCard64 (lambda);  WriteLn;

        (* Solve d*e mod lambda = 1 for d. *)

        (*d := EEA2 (e, lambda);*)
        EEA1b (lambda, e, d);

        WriteString ("The public key is (n = ");  WriteCard64(n);
        WriteString (", e = ");  WriteCard (e);  WriteChar (")");  WriteLn;

        WriteString ("The private key is (n = ");  WriteCard64(n);
        WriteString (", d = ");  WriteInt (d);  WriteChar (")");  WriteLn;

    END RSA_Generate_example;

(************************************************************************)
(*                              TEST FUNCTIONS                          *)
(************************************************************************)

PROCEDURE BasicTests;

    VAR res: ARRAY [0..2] OF INTEGER;
        a, b, e, m: CARDINAL;

    BEGIN
        a := 75;  b := 90;
        WriteString ("GCD of ");  WriteCard (a);
        WriteString (" and ");  WriteCard (b);
        WriteString (" is ");  WriteCard (gcd (a, b));
        WriteLn;

        WriteString ("Extended GCD (");  WriteCard (a);
        WriteString (", ");  WriteCard (b);
        WriteString (") gives [");
        xgcd (res, a, b);
        WriteInt (res[0]);  WriteString ("  ");
        WriteInt (res[1]);  WriteString ("  ");
        WriteInt (res[2]);  WriteString ("]");  WriteLn;

        (*
        WriteString ("LCM of ");  WriteCard (a);
        WriteString (" and ");  WriteCard (b);
        WriteString (" is ");  WriteCard (lcm (a, b));
        WriteLn;
        *)

        a := 40000;  b := 30000;
        WriteString ("Product of ");  WriteCard (a);
        WriteString (" and ");  WriteCard (b);
        WriteString (" is ");  WriteCard64 (Mul32 (a, b));
        WriteLn;

        b := 4;  e := 13;  m := 497;
        WriteCard (b);  WriteChar ('^');  WriteCard (e);
        WriteString (" MOD ");  WriteCard (m);
        WriteString (" = ");  WriteCard (ModularPower (b, e, m));
        WriteLn;

        WriteString ("A random prime is ");  WriteCard (RandomPrime());
        WriteLn;
        WriteString ("Another random prime is ");  WriteCard (RandomPrime());
        WriteLn;

    END BasicTests;

(************************************************************************)

PROCEDURE EEAtest;

    (* Testing extended Euclid algorithm. *)

    CONST
        a = 240;
        b = 46;

    VAR m, x, y, val: INTEGER;

    BEGIN
        (*
        WriteString ("EEA test with a = ");  WriteCard (a);
        WriteString (", ");  WriteCard (b);  WriteLn;
        EEA (a, b, x, y);
        WriteString ("Bezout coefficients are ");  WriteInt (x);
        WriteString (", ");  WriteInt (y);  WriteLn;
        *)

        m := b;
        WriteString ("EEA1 test with a = ");  WriteCard (a);
        WriteString (", m = ");  WriteCard (m);  WriteLn;
        EEA1b (CARD64{0,a}, m, val);
        WriteString ("ax mod m = gcd(a,m) has solution x = ");  WriteInt (val);  WriteLn;
        WriteString ("Check: ax = ");  WriteCard (a*val);  WriteLn;
        WriteString ("ax mod m = ");  WriteCard (a*val MOD m);  WriteLn;

    END EEAtest;

(************************************************************************)

PROCEDURE PrimeTest;

    (* We generate a list of certain primes using the sieve of          *)
    (* Eratosthenes, and a list of probably primes using procedure      *)
    (* ProbablePrime, so we can compare the two lists.                  *)

    CONST max = 5000;

    VAR list1, list2: ARRAY [1..max] OF CARDINAL;
        j, k, p, step, pos: CARDINAL;
        found: BOOLEAN;

    BEGIN
        (* Fill list1 with all candidates. *)

        FOR j := 1 TO max DO
            list1[j] := j;
        END (*FOR*);

        (* Sieve out the composites. *)

        p := 2;
        WHILE p*p <= max DO
            step := p;  pos := p + step;
            WHILE pos <= max DO
                list1[pos] := 0;
                INC (pos, p);
            END (*WHILE*);
            REPEAT
               INC (p);
            UNTIL (p*p > max) OR (list1[p] <> 0);
        END (*WHILE*);

        (* Compact the list. *)

        p := 4;  pos := 4;
        LOOP
            IF p > max THEN EXIT(*LOOP*) END(*IF*);
            WHILE (pos <= max) AND (list1[pos] = 0) DO
                INC (pos);
            END (*WHILE*);
            IF pos > max THEN EXIT(*LOOP*) END (*IF*);
            list1[p] := list1[pos];
            INC (p);  INC(pos);
        END (*LOOP*);

        (* Fill list2. *)

        pos := 1;  list2[pos] := 0;
        FOR p := 1 TO max DO
            IF ProbablePrime(p) THEN
                list2[pos] := p;
                INC (pos);
            END (*IF*);
        END (*FOR*);

        (* Write out list1. *)

        (*
        WriteString ("From sieve of Eratosthenes:");  WriteLn;
        pos := 1;
        WHILE list1[pos] <> 0 DO
            WriteCard (list1[pos]);  WriteChar (' ');
            INC (pos);
        END (*WHILE*);
        WriteLn;
        *)

        (* Write out list2. *)

        (*
        WriteString ("From ProbablePrime test:");  WriteLn;
        pos := 1;
        WHILE list2[pos] <> 0 DO
            WriteCard (list2[pos]);  WriteChar (' ');
            INC (pos);
        END (*WHILE*);
        WriteLn;
        *)

        (* Check whether list2 has any entries that are not in list1. *)

        WriteString ("Pseudoprimes found: ");
        j := 1;  k := 1;  found := FALSE;
        WHILE list2[k] <> 0 DO
            WHILE (list1[j] <> 0) AND (list1[j] < list2[k]) DO
                INC (j);
            END (*WHILE*);
            IF list1[j] > list2[k] THEN
                WriteCard (list2[k]);  WriteChar (' ');
                found := TRUE;
            END (*IF*);
            INC (k);
        END (*WHILE*);
        IF NOT found THEN
            WriteString ("(none)");
        END (*IF*);
        WriteLn;

    END PrimeTest;

(************************************************************************)

BEGIN
    EEAtest;
    (*
    Randomize;
    BasicTests;
    PrimeTest;
    RSA_Generate_example;
    *)
END draft.

