MODULE draft1;

        (********************************************************)
        (*                                                      *)
        (*         For trying out miscellaneous ideas.          *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Last edited:        22 March 2018                   *)
        (*  Status:             Work in progress                *)
        (*                                                      *)
        (********************************************************)


FROM SYSTEM IMPORT CAST, CARD32;

IMPORT BigNum;

FROM BigNum IMPORT
    (* type *)  BN,
    (* proc *)  ShowBNUsage, WriteBignum, ModularPower;

FROM Arith64 IMPORT
    (* const*)  Zero64,
    (* type *)  CARD64, INT64,
    (* proc *)  Mul32, Div10, LongDiv64;

FROM RandCard IMPORT
    (* proc *)  Randomize;

FROM Primes IMPORT
    (* proc *)  RandomPrime, Jacobi, IsPerfectSquare;

FROM STextIO IMPORT
    (* proc *)  WriteChar, WriteString, WriteLn;

(************************************************************************)
(*                           MODULAR ARITHMETIC                         *)
(************************************************************************)

PROCEDURE gcd (a, b: BN): BN;

    (* Greatest common divisor. *)

    VAR a1, b1, temp, Q, R: BN;

    BEGIN
        (*WriteString ("Entering gcd, ");  ShowBNUsage;*)
        a1 := BigNum.CopyBN (a);
        IF BigNum.Cmp(a, b) = 0 THEN RETURN a1;
        ELSE
            b1 := BigNum.CopyBN(b);
            IF BigNum.Cmp (b1, a1) > 0 THEN
                temp := b1;  b1 := a1;  a1 := temp;
            END (*IF*);

            (* Now a1 > b1 for the rest of the calculation. *)

            WHILE NOT BigNum.IsZero(b1) DO
                BigNum.Divide (a1, b1, Q, R);
                BigNum.Discard (Q);
                BigNum.Discard (a1);
                a1 := b1;
                b1 := R;
            END (*WHILE*);
        END (*IF*);
        BigNum.Discard (b1);
        (*WriteString ("Leaving gcd, ");  ShowBNUsage;*)
        RETURN a1;
    END gcd;

(************************************************************************)

PROCEDURE inverse (a: CARDINAL;  m: CARD64): CARD64;

    (* Returns x such that ax = 1 mod m.  This will work only if a and  *)
    (* m are relatively prime.                                          *)

    (* The extended Euclid algorithm computes a triple (g, x, y) such   *)
    (* that ax + my = g, and g = gcd(a,m).  This modified version       *)
    (* returns only x, since all we want is a solution of ax = g mod m. *)

    VAR t0, t1, q: INTEGER;
        s0, s1: INT64;
        x, originalm: CARD64;

    BEGIN
        (*
        originalm := m;
        s0 := INT64{0,1};  s1 := INT64{0,0};
        t0 := 0;  t1 := 1;
        LOOP
            q := a DIV m;  a := a MOD m;
            DEC (s0, q*s1);
            DEC (t0, q*t1);
            IF a = 0 THEN
                x := s1;
                EXIT (*LOOP*);
            END (*IF*);
            q := m DIV a;  m := m MOD a;
            DEC (s1, q*s0);
            DEC (t1, q*t0);
            IF m = 0 THEN
                x := s0;
                EXIT (*LOOP*);
            END (*IF*);
        END (*LOOP*);
        IF x < 0 THEN INC (x, originalm);
        END (*IF*);
        *)
        x.high := 0;
        x.low := 1;
        RETURN x;
    END inverse;

(************************************************************************)

PROCEDURE xgcd (a, b: BN;  VAR (*OUT*) x, y: BN);

    (* Extended Euclid algorithm, an implementation from first          *)
    (* principles.  We calculate x and y such that ax+by = gcd(a,b).    *)

    VAR Q, oldR, R, newR, temp: BN;
        xold, xnew: BN;
        yold, ynew: BN;

    BEGIN
        (*WriteString ("Entering xgcd, ");  ShowBNUsage;*)
        oldR := BigNum.CopyBN(a);  R := BigNum.CopyBN(b);
        xold := BigNum.MakeBignum(1);  x := BigNum.MakeBignum(0);
        yold := BigNum.MakeBignum(0);  y := BigNum.MakeBignum(1);
        LOOP
            BigNum.Divide (oldR, R, Q, newR);
            temp := BigNum.Prod (Q, x);
            xnew := BigNum.Diff (xold, temp);
            BigNum.Discard (temp);
            temp := BigNum.Prod (Q, y);
            ynew := BigNum.Diff (yold, temp);
            BigNum.Discard (temp);
            BigNum.Discard (Q);
            IF BigNum.IsZero(newR) THEN EXIT(*LOOP*) END(*LOOP*);
            BigNum.Discard (oldR);  oldR := R;  R := newR;
            BigNum.Discard (xold);  xold := x;  x := xnew;
            BigNum.Discard (yold);  yold := y;  y := ynew;
        END (*LOOP*);
        BigNum.Discard (Q);  BigNum.Discard (oldR);
        BigNum.Discard (R);  BigNum.Discard (newR);
        BigNum.Discard (xold);  BigNum.Discard (xnew);
        BigNum.Discard (yold);  BigNum.Discard (ynew);
        (*WriteString ("Leaving xgcd, ");  ShowBNUsage;*)
    END xgcd;

(************************************************************************)

PROCEDURE exgcd (a, m: BN): BN;

    (* A variant of xgcd where we don't care about the solution for y.  *)
    (* This is of interest mostly in the case where a and m are         *)
    (* relatively prime, because in that case x = inverse(a) mod m.     *)

    VAR Q, oldR, R, newR, temp: BN;
        xold, x, xnew: BN;

    BEGIN
        (*WriteString ("Entering exgcd, ");  ShowBNUsage;*)
        oldR := BigNum.CopyBN(a);  R := BigNum.CopyBN(m);
        xold := BigNum.MakeBignum(1);  x := BigNum.MakeBignum(0);
        LOOP
            BigNum.Divide (oldR, R, Q, newR);
            temp := BigNum.Prod (Q, x);
            xnew := BigNum.Diff (xold, temp);
            BigNum.Discard (temp);
            BigNum.Discard (Q);
            IF BigNum.IsZero(newR) THEN EXIT(*LOOP*) END(*LOOP*);
            BigNum.Discard (oldR);  oldR := R;  R := newR;
            BigNum.Discard (xold);  xold := x;  x := xnew;
        END (*LOOP*);
        BigNum.Discard (Q);  BigNum.Discard (oldR);
        BigNum.Discard (R);  BigNum.Discard (newR);
        BigNum.Discard (xold);  BigNum.Discard (xnew);
        (*WriteString ("Leaving exgcd, ");  ShowBNUsage;*)
        RETURN x;
    END exgcd;

(************************************************************************)

PROCEDURE Inverse (a, m: BN): BN;

    (* Returns inverse(a) mod m.                    *)
    (* Assumption: a and m are relatively prime.    *)

    VAR Q, oldR, R, newR, temp: BN;
        xold, x, xnew: BN;
        aint, bint: INTEGER;

    BEGIN
        RETURN exgcd (a, m);
    END Inverse;

(************************************************************************)

PROCEDURE lcm (a, b: BN): BN;

    (* Least common multiple. *)

    VAR c, d: BN;

    BEGIN
        (*WriteString ("Entering lcm, ");  ShowBNUsage;*)
        IF a = b THEN
            (*WriteString ("Leaving lcm, ");  ShowBNUsage;*)
            RETURN BigNum.CopyBN (a);
        ELSE
            c := gcd(a,b);
            d := BigNum.Prod (c, b);
            BigNum.Discard (c);
            (*WriteString ("Leaving lcm, ");  ShowBNUsage;*)
            RETURN d;
        END (*IF*);
    END lcm;

(************************************************************************)

PROCEDURE iJacobi (a, b: INTEGER): INTEGER;

    (* Calculates the Jacobi symbol (a/b).   Integer version.  *)

    VAR j, temp: INTEGER;

    BEGIN
        IF (b <= 0) OR NOT ODD(b) THEN RETURN 0; END(*IF*);

        j := 1;
        IF a < 0 THEN
            a := -a;
            IF b MOD 4 = 3 THEN j := -j END (*IF*);
        END (*IF*);

        WHILE a <> 0 DO
           WHILE NOT ODD(a) DO

              (* Process factors of 2: (2,b) = -1 if b = 3 or 5 (mod 8) *)

              a := a DIV 2;
              IF (b MOD 8 = 3) OR (b MOD 8 = 5) THEN j := -j END (*IF*);

           END (*WHILE*);

           (* Quadratic reciprocity: (a,b) = (b,a) if either a or b is  *)
           (* equal to 1 (mod 4), and(a,b) = -(b,a) if both a and b are *)
           (* equal to 3 (mod 4).  The other cases were eliminated      *)
           (* above when we handled the case of even a.                 *)

           temp := a;  a := b;  b := temp;
           IF (a MOD 4 = 3) AND (b MOD 4 = 3) THEN j := -j END (*IF*);
           a := a MOD b;

        END (*WHILE*);

        IF b = 1 THEN RETURN j
        ELSE RETURN 0
        END (*IF*);

    END iJacobi;

(************************************************************************)
(*                             RSA EXAMPLE                              *)
(************************************************************************)

(**)
PROCEDURE RSA_Generate_example;

    (* This uses artificially small numbers, to let me see what is      *)
    (* going on without having to deal with the complexities of very    *)
    (* wide numbers.                                                    *)

    CONST nwords = 1;

    VAR p, q, pm1, qm1, n, lambda, e, d, Q, R: BN;

    BEGIN
        WriteLn;  WriteString ("RSA example:");  WriteLn;

        (* Choose two distinct prime numbers p and q.  Let n = p*q *)

        p := RandomPrime(nwords);  q := RandomPrime(nwords);
        WriteString ("We have chosen two random primes, p = ");
        BigNum.WriteBignum (p);  WriteString (" and q = ");  BigNum.WriteBignum (q);
        WriteLn;
        n := BigNum.Prod (p, q);

        pm1 := BigNum.CopyBN(p);
        BigNum.Decr (pm1);
        qm1 := BigNum.CopyBN(q);
        BigNum.Decr (qm1);

        lambda := lcm (pm1, qm1);

        (* Choose e such that 1 < e < lambda that is coprime to lambda. *)
        (* Making e prime simplifies the coprime check.                 *)

        e := BigNum.MakeBignum (17);
        BigNum.Divide (lambda, e, Q, R);
        IF BigNum.IsZero(R) THEN
            WriteString ("Poorly chosen e. Try again.");  WriteLn;
        END (*IF*);

        WriteString ("lambda = ");  BigNum.WriteBignum (lambda);  WriteLn;

        (* Solve d*e mod lambda = 1 for d. *)

        d := exgcd (e, lambda);
        (*EEA1b (lambda, e, d);*)

        WriteString ("The public key is (n = ");  BigNum.WriteBignum(n);
        WriteString (", e = ");  BigNum.WriteBignum (e);  WriteChar (")");  WriteLn;

        WriteString ("The private key is (n = ");  BigNum.WriteBignum (n);
        WriteString (", d = ");  BigNum.WriteBignum (d);  WriteChar (")");  WriteLn;

    END RSA_Generate_example;
(**)

(************************************************************************)
(*                             TEST FUNCTIONS                           *)
(************************************************************************)

PROCEDURE OldSquareTest;

    (********************************************************************)

    PROCEDURE TestN (N: CARDINAL);

        VAR B: BN;

        BEGIN
            B := BigNum.MakeBignum (N);
            BigNum.WriteBignum (B);
            IF IsPerfectSquare (B) THEN
                WriteString (" is a perfect square");
            ELSE
                WriteString (" is not a perfect square");
            END (*IF*);
            WriteLn;
        END TestN;

    (********************************************************************)

    VAR j: CARDINAL;

    BEGIN
        FOR j := 1 TO 10 DO
            TestN (j);
        END (*FOR*);
        FOR j := 16777216 TO 20000000 BY 8193 DO
            TestN (j);
        END (*FOR*);
    END OldSquareTest;

(************************************************************************)

PROCEDURE SquareTest;

    VAR N1, N2, N1sq, N2sq, N, count, expected: BN;

    BEGIN
        ShowBNUsage;
        N1 := BigNum.MakeBignum(130);
        N1sq := BigNum.Prod (N1, N1);
        N2 := BigNum.MakeBignum(1000);
        N2sq := BigNum.Prod (N2, N2);
        N := BigNum.Diff (N2, N1);
        BigNum.Incr (N);
        WriteString ("There should be ");
        BigNum.WriteBignum (N);
        WriteString (" squares between ");
        BigNum.WriteBignum (N1sq);
        WriteString (" and ");
        BigNum.WriteBignum (N2sq);
        WriteLn;
        expected := N;
        BigNum.Discard (N1);
        BigNum.Discard (N2);
        N := N1sq;
        count := BigNum.MakeBignum(0);
        WHILE BigNum.Cmp(N, N2sq) <= 0 DO
            IF IsPerfectSquare (N) THEN
                (**)
                WriteChar (" ");
                BigNum.WriteBignum (N);
                (**)
                BigNum.Incr(count);
            END (*IF*);
            BigNum.Incr(N);
        END (*WHILE*);
        WriteLn;
        WriteString ("      I expected ");
        BigNum.WriteBignum (expected);
        WriteString (", and counted ");
        BigNum.WriteBignum (count);
        WriteLn;
        BigNum.Discard (N);
        BigNum.Discard (N2sq);
        BigNum.Discard (count);
        BigNum.Discard (expected);
        ShowBNUsage;
    END SquareTest;

(************************************************************************)

PROCEDURE BasicTests;

    VAR a, b, m, x, y, e: BN;

    BEGIN
        WriteString ("Entering BasicTests, ");  ShowBNUsage;
        a := BigNum.MakeBignum (75);  b := BigNum.MakeBignum (90);
        y := gcd (a, b);
        WriteString ("GCD of ");  WriteBignum (a);
        WriteString (" and ");  WriteBignum (b);
        WriteString (" is ");
        WriteBignum (y);
        WriteLn;
        BigNum.Discard (y);

        xgcd (a, b, x, y);
        WriteString ("Extended GCD (");  WriteBignum (a);
        WriteString (", ");  WriteBignum (b);
        WriteString (") gives x = ");
        WriteBignum (x);  WriteString (",  y = ");
        WriteBignum (y);  WriteLn;
        BigNum.Discard (x);  BigNum.Discard (y);

        x := exgcd (a, b);
        WriteString ("Double extended GCD (");  WriteBignum (a);
        WriteString (", ");  WriteBignum (b);
        WriteString (") gives x = ");
        WriteBignum (x);  WriteLn;
        BigNum.Discard (x);

        y := lcm (a, b);
        WriteString ("LCM of ");  WriteBignum (a);
        WriteString (" and ");  WriteBignum (b);
        WriteString (" is ");
        WriteBignum (y);
        WriteLn;
        BigNum.Discard (y);
        BigNum.Discard (a);  BigNum.Discard (b);

        (* Modular power.  *)

        b := BigNum.MakeBignum(4);  e := BigNum.MakeBignum(13);  m := BigNum.MakeBignum(497);
        WriteBignum (b);  WriteChar ('^');  WriteBignum (e);
        WriteString (" MOD ");  WriteBignum (m);
        WriteString (" = ");
        y := ModularPower (b, e, m);
        WriteBignum (y);
        WriteLn;
        BigNum.Discard (y);
        BigNum.Discard (m);
        BigNum.Discard (e);
        BigNum.Discard (b);
        WriteString ("Leaving BasicTests, ");  ShowBNUsage;

    END BasicTests;

(************************************************************************)

PROCEDURE PrimeTest;

    (* Random prime. *)

    VAR y: BN;

    BEGIN
        WriteString ("Entering PrimeTest, ");  ShowBNUsage;
        y := RandomPrime(1);
        WriteString ("A random prime is ");
        WriteBignum (y);
        BigNum.Discard (y);
        WriteLn;
        y := RandomPrime(3);
        WriteString ("Another random prime is ");
        WriteBignum (y);
        BigNum.Discard (y);
        WriteLn;
        WriteString ("Partway through PrimeTest, ");  ShowBNUsage;
        y := RandomPrime(5);
        WriteString ("Another random prime is ");
        WriteBignum (y);
        BigNum.Discard (y);
        WriteLn;
        WriteString ("Leaving PrimeTest, ");  ShowBNUsage;

    END PrimeTest;

(************************************************************************)

PROCEDURE EEAtest;

    (* Testing extended Euclid algorithm. *)

    CONST
        aval = 240;
        bval = 46;

    VAR a, b, x, y, prod1, prod2, sum, m, val: BN;

    BEGIN
        ShowBNUsage;
        a := BigNum.MakeBignum (aval);
        b := BigNum.MakeBignum (bval);
        WriteString ("EEA test with a = ");  WriteBignum (a);
        WriteString (", b = ");  WriteBignum (b);  WriteLn;
        xgcd (a, b, x, y);
        WriteString ("Bezout coefficients are ");  WriteBignum (x);
        WriteString (", ");  WriteBignum (y);  WriteLn;
        prod1 := BigNum.Prod (a, x);
        prod2 := BigNum.Prod (b, y);
        BigNum.Discard (x);  BigNum.Discard (y);
        sum := BigNum.Sum (prod1, prod2);
        WriteString ("ax + by = ");  WriteBignum (sum);  WriteLn;
        BigNum.Discard (prod1);  BigNum.Discard (prod2);  BigNum.Discard (sum);

        m := BigNum.CopyBN (b);
        WriteString ("EEA1 test with a = ");  WriteBignum (a);
        WriteString (", m = ");  WriteBignum (m);  WriteLn;
        val := exgcd (a, m);
        WriteString ("ax mod m = gcd(a,m) has solution x = ");  WriteBignum (val);  WriteLn;
        WriteString ("Check: ax = ");  prod1 := BigNum.Prod(a, val);
        BigNum.Discard (val);
        WriteBignum (prod1);  WriteLn;
        BigNum.Divide (prod1, m, prod2, val);
        BigNum.Discard (prod1);
        BigNum.Discard (prod2);
        WriteString ("ax mod m = ");  WriteBignum (val);  WriteLn;
        BigNum.Discard (val);
        BigNum.Discard (m);

        BigNum.Discard (b);  BigNum.Discard (a);
        ShowBNUsage;
    END EEAtest;

(************************************************************************)

PROCEDURE JacobiTest;

    (* Tests the Jacobi function.  *)

    VAR a, b, result: INTEGER;
        k, n: BN;

    BEGIN
        (*WriteString ("Entering JacobiTest, ");  ShowBNUsage;*)
        WriteString ("Table of Jacobi function (a/b).");  WriteLn;
        FOR b := 1 TO 21 BY 2 DO
            FOR a := 1 TO 21 DO
                result := iJacobi (a,b);
                IF result = -1 THEN WriteString(" -1")
                ELSIF result = 0 THEN WriteString("  0")
                ELSIF result = 1 THEN WriteString("  1")
                ELSE
                    WriteString ("Something has gone wrong in Jacobi function");
                END (*IF*);
            END (*FOR*);
            WriteLn;
        END (*FOR*);

        (* Now repeat the test for the Bignum version. *)

        WriteString ("Bignum version.");  WriteLn;
        FOR b := 1 TO 21 BY 2 DO
            n := BigNum.MakeBignum (b);
            FOR a := 1 TO 21 DO
                k := BigNum.MakeBignum (a);
                result := Jacobi (k,n);
                IF result = -1 THEN WriteString(" -1")
                ELSIF result = 0 THEN WriteString("  0")
                ELSIF result = 1 THEN WriteString("  1")
                ELSE
                    WriteString ("Something has gone wrong in Jacobi function");
                END (*IF*);
                BigNum.Discard (k);
            END (*FOR*);
            WriteLn;
            BigNum.Discard (n);
        END (*FOR*);

        (*WriteString ("Leaving JacobiTest, ");  ShowBNUsage;*)
    END JacobiTest;

(************************************************************************)

BEGIN
    (*
    EEAtest;
    Randomize;
    BasicTests;
    PrimeTest;
    SquareTest;
    JacobiTest;
    *)
    (**)
    RSA_Generate_example;
    (**)
END draft1.

