(**************************************************************************)
(*                                                                        *)
(*  SHA512 encryption                                                     *)
(*  Copyright (C) 2017   Peter Moylan                                     *)
(*                                                                        *)
(*  This program is free software: you can redistribute it and/or modify  *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 3 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  This program is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *)
(*                                                                        *)
(*  To contact author:   http://www.pmoylan.org   peter@pmoylan.org       *)
(*                                                                        *)
(**************************************************************************)

MODULE TestSHA512;

        (********************************************************)
        (*                                                      *)
        (*         TEST OF SHA-384 AND SHA-512 ENCRYPTION       *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            1 October 2017                  *)
        (*  Last edited:        4 October 2017                  *)
        (*  Status:             Working                         *)
        (*                                                      *)
        (* I am using examples taken from                       *)
        (*     https://www.di-mgt.com.au/sha_testvectors.html   *)
        (* Later I will rely on the HMAC tests of RFC 4231.     *)
        (*                                                      *)
        (********************************************************)


IMPORT STextIO;

FROM SHA512 IMPORT
    (* type *)  SHA512_CTX,
    (* proc *)  SHA384Init, SHA512Init, SHA512Update, SHA512Final,
                SHA512FinalDigest;

(************************************************************************)

PROCEDURE DoTest (variant: CARDINAL;  data: ARRAY OF CHAR);

    VAR ctx: SHA512_CTX;
        result: ARRAY [0..255] OF CHAR;        (* allow for some spaces *)

    BEGIN
        IF variant = 384 THEN
            ctx := SHA384Init();
        ELSE
            ctx := SHA512Init();
        END (*IF*);
        SHA512Update (ctx, data, LENGTH(data));
        SHA512FinalDigest (ctx, result);
        IF variant = 384 THEN
            STextIO.WriteString ('SHA384(');
        ELSE
            STextIO.WriteString ('SHA512(');
        END (*IF*);
        STextIO.WriteString (data);
        STextIO.WriteString (') = ');  STextIO.WriteLn;
        STextIO.WriteString ('    ');
        STextIO.WriteString (result);
        STextIO.WriteLn;
    END DoTest;

(************************************************************************)

PROCEDURE MillionTest (variant: CARDINAL);

    (* One million repetitions of the character 'a'. *)

    VAR ctx: SHA512_CTX;
        result: ARRAY [0..255] OF CHAR;        (* allow for some spaces *)
        j: CARDINAL;

    BEGIN
        IF variant = 384 THEN
            ctx := SHA384Init();
        ELSE
            ctx := SHA512Init();
        END (*IF*);
        FOR j := 1 TO 1000000 DO
            SHA512Update (ctx, 'a', 1);
        END (*FOR*);
        SHA512FinalDigest (ctx, result);
        IF variant = 384 THEN
            STextIO.WriteString ('SHA384');
        ELSE
            STextIO.WriteString ('SHA512');
        END (*IF*);
        STextIO.WriteString ('(10^6 x a) = ');  STextIO.WriteLn;
        STextIO.WriteString ('    ');
        STextIO.WriteString (result);
        STextIO.WriteLn;
    END MillionTest;

(************************************************************************)

PROCEDURE TestData (data: ARRAY OF CHAR);

    BEGIN
        DoTest (384, data);
        DoTest (512, data);
    END TestData;

(************************************************************************)

PROCEDURE RunTheTest;

    (* These tests are taken from                               *)
    (*      https://www.di-mgt.com.au/sha_testvectors.html      *)

    BEGIN
        (* Test with no data. *)
        TestData ("");

        (* Some examples taken from https://www.di-mgt.com.au/sha_testvectors.html *)

        TestData ("abc");

        TestData ("abcdbcdecdefdefgefghfghighijhijkijkljklmklmnlmnomnopnopq");

        TestData ("abcdefghbcdefghicdefghijdefghijkefghijklfghijklmghijklmnhijklmnoijklmnopjklmnopqklmnopqrlmnopqrsmnopqrstnopqrstu");

        MillionTest (384);
        MillionTest (512);

END RunTheTest;

(************************************************************************)

BEGIN
    RunTheTest;
END TestSHA512.
