(**************************************************************************)
(*                                                                        *)
(*  SHA256 encryption                                                     *)
(*  Copyright (C) 2016   Peter Moylan                                     *)
(*                                                                        *)
(*  This program is free software: you can redistribute it and/or modify  *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 3 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  This program is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *)
(*                                                                        *)
(*  To contact author:   http://www.pmoylan.org   peter@pmoylan.org       *)
(*                                                                        *)
(**************************************************************************)

MODULE TestSHA2;

        (********************************************************)
        (*                                                      *)
        (*         TEST OF SHA-224 AND SHA-256 ENCRYPTION       *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            14 December 2016                *)
        (*  Last edited:        29 September 2017               *)
        (*  Status:                                             *)
        (*      This is by no means a complete suite of tests,  *)
        (*      but the results so far are good.                *)
        (*                                                      *)
        (********************************************************)


IMPORT STextIO;

FROM SHA2 IMPORT
    (* type *)  SHA2_CTX,
    (* proc *)  SHA224Init, SHA256Init, SHA2Update, SHA2Final,
                SHA2FinalDigest;

(************************************************************************)

PROCEDURE DoTest (variant: CARDINAL;  data: ARRAY OF CHAR);

    VAR ctx: SHA2_CTX;
        result: ARRAY [0..63] OF CHAR;

    BEGIN
        IF variant = 224 THEN
            ctx := SHA224Init();
        ELSE
            ctx := SHA256Init();
        END (*IF*);
        SHA2Update (ctx, data, LENGTH(data));
        SHA2FinalDigest (ctx, result);
        IF variant = 224 THEN
            STextIO.WriteString ('SHA224(');
        ELSE
            STextIO.WriteString ('SHA256(');
        END (*IF*);
        STextIO.WriteString (data);
        STextIO.WriteString (') = ');  STextIO.WriteLn;
        STextIO.WriteString ('    ');
        STextIO.WriteString (result);
        STextIO.WriteLn;
    END DoTest;

(************************************************************************)

PROCEDURE TestData (data: ARRAY OF CHAR);

    BEGIN
        DoTest (224, data);
        DoTest (256, data);
    END TestData;

(************************************************************************)

PROCEDURE RunTheTest;

    BEGIN
        (* Test with no data. *)
        TestData ("");
        (* SHA-224 result should be  *)
        (* d14a028c2a3a2bc9476102bb288234c415a2b01f828ea62ac5b3e42f *)
        (* SHA-256 result should be  *)
        (* e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855 *)

        (* Test with a non-empty string. *)
        TestData ("The quick brown fox jumps over the lazy dog");
        (* SHA-224 result should be  *)
        (* 730e109bd7a8a32b1cb9d9a09aa2325d2430587ddbc0c38bad911525 *)
        (* SHA-256 result should be  *)
        (* D7A8FBB307D7809469CA9ABCB0082E4F8D5651E46D3CDB762D02D0BF37C9E592 *)

        (* A slightly different string. *)
        TestData ("The quick brown fox jumps over the lazy dog.");
        (* SHA-224 result should be  *)
        (* 619cba8e8e05826e9b8c519c0a5c68f4fb653e8a3d8aa04bb2c8cd4c *)
        (* SHA-256 result should be  *)
        (* EF537F25C895BFA782526529A9B63D97AA631564D5D789C2B765448C8635FB6C *)

        (* Some examples taken from https://www.di-mgt.com.au/sha_testvectors.html *)

        TestData ("abc");
        (* SHA-224 result should be  *)
        (* 23097d22 3405d822 8642a477 bda255b3 2aadbce4 bda0b3f7 e36c9da7 *)
        (* SHA-256 result should be  *)
        (* ba7816bf 8f01cfea 414140de 5dae2223 b00361a3 96177a9c b410ff61 f20015ad *)

        TestData ("abcdbcdecdefdefgefghfghighijhijkijkljklmklmnlmnomnopnopq");
        (* SHA-224 result should be  *)
        (* 75388b16 512776cc 5dba5da1 fd890150 b0c6455c b4f58b19 52522525 *)
        (* SHA-256 result should be  *)
        (* 248d6a61 d20638b8 e5c02693 0c3e6039 a33ce459 64ff2167 f6ecedd4 19db06c1 *)

        TestData ("abcdefghbcdefghicdefghijdefghijkefghijklfghijklmghijklmnhijklmnoijklmnopjklmnopqklmnopqrlmnopqrsmnopqrstnopqrstu");
        (* SHA-224 result should be  *)
        (* c97ca9a5 59850ce9 7a04a96d ef6d99a9 e0e0e2ab 14e6b8df 265fc0b3 *)
        (* SHA-256 result should be  *)
        (* cf5b16a7 78af8380 036ce59e 7b049237 0b249b11 e8f07a51 afac4503 7afee9d1 *)

END RunTheTest;

(************************************************************************)

BEGIN
    RunTheTest;
END TestSHA2.
