MODULE testRSA;

        (********************************************************)
        (*                                                      *)
        (*          Test of RSA encryption and decryption       *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Last edited:        6 May 2018                      *)
        (*  Status:             Work in progress                *)
        (*                                                      *)
        (*      In case you're wondering what RSA stands for,   *)
        (*      it's Rivest-Shamir-Adleman, after the           *)
        (*      inventors of the algorithm.                     *)
        (*                                                      *)
        (********************************************************)


IMPORT BigNum;

FROM RSA IMPORT
    (* proc *)  Encrypt, Decrypt;

FROM RSAKeys IMPORT
    (* type *)  RSAKeyType, DiscardKey;

FROM RSAGenerate IMPORT
    (* proc *)  RSA_Generate;

FROM BigNum IMPORT
    (* type *)  BN;

FROM STextIO IMPORT
    (* proc *)  WriteChar, WriteString, WriteLn;

(************************************************************************)
(*                             TEST FUNCTIONS                           *)
(************************************************************************)

PROCEDURE MainTest;

    CONST keylength = 16;

    VAR key: RSAKeyType;
        message, ciphertext, result: BN;

    BEGIN
        WriteString ("Generating a key, please be patient");  WriteLn;
        key := RSA_Generate(keylength);

        WriteString ("The public key is");  WriteLn;
        WriteString ("n ");  BigNum.WriteBignum(key.n);  WriteLn;
        WriteString ("e ");  BigNum.WriteBignum (key.public);  WriteLn;

        WriteString ("The private key is");  WriteLn;
        WriteString ("n ");  BigNum.WriteBignum (key.n);  WriteLn;
        WriteString ("d ");  BigNum.WriteBignum (key.private);  WriteLn;

        WriteString ("ENCRYPTION EXAMPLE");  WriteLn;
        message := BigNum.Rand(keylength - 1);
        WriteString ("text ");  BigNum.WriteBignum(message);
        WriteLn;
        ciphertext := Encrypt (message, key);
        WriteString ("encr ");  BigNum.WriteBignum(ciphertext);
        WriteLn;
        result := Decrypt (ciphertext, key);
        WriteString ("decr ");  BigNum.WriteBignum(result);
        WriteLn;
        BigNum.Discard (result);
        BigNum.Discard (ciphertext);
        BigNum.Discard (message);

        DiscardKey (key);

    END MainTest;

(************************************************************************)

PROCEDURE SimpleTest;

    (* Example taken from the web, so that I can compare the numbers. *)

    VAR key: RSAKeyType;
        message, ciphertext, result: BN;

    BEGIN
        WriteString ("Simple test");  WriteLn;

        key.n := BigNum.MakeBignum (3233);
        key.public := BigNum.MakeBignum (17);
        key.private := BigNum.MakeBignum (413);
        key.p := BigNum.MakeBignum (61);
        key.q := BigNum.MakeBignum (53);
        key.dp := BigNum.MakeBignum (53);
        key.dq := BigNum.MakeBignum (49);
        key.qinv := BigNum.MakeBignum (38);

        WriteString ("The public key is (n = ");  BigNum.WriteBignum(key.n);
        WriteString (", e = ");  BigNum.WriteBignum (key.public);  WriteChar (")");  WriteLn;

        WriteString ("The private key is (n = ");  BigNum.WriteBignum (key.n);
        WriteString (", d = ");  BigNum.WriteBignum (key.private);  WriteChar (")");  WriteLn;

        message := BigNum.MakeBignum(65);
        WriteString ("message = ");  BigNum.WriteBignum(message);
        WriteLn;
        ciphertext := Encrypt (message, key);
        WriteString ("ciphertext = ");  BigNum.WriteBignum(ciphertext);
        WriteLn;
        result := Decrypt (ciphertext, key);
        WriteString ("decrypted message = ");  BigNum.WriteBignum(result);
        WriteLn;
        BigNum.Discard (result);
        BigNum.Discard (ciphertext);
        BigNum.Discard (message);
        DiscardKey (key);
        WriteString ("========================================");  WriteLn;

    END SimpleTest;

(************************************************************************)

BEGIN
    SimpleTest;
    MainTest;
END testRSA.

