MODULE MakeTable4x16;

    (********************************************************)
    (* Reformats some data from NIST.FIPS.197 to make it    *)
    (* more useable.                                        *)
    (********************************************************)


IMPORT TextIO, SeqFile;

FROM IOChan IMPORT ChanId;

(****************************************************************)

CONST
    InFileName = "Indata.txt";
    OutFileName = "outdata.txt";

TYPE
    CharSet = SET OF CHAR;

VAR Data: ARRAY [0..3],[0..15] OF CARDINAL;

(****************************************************************)
(*                           INPUT                              *)
(****************************************************************)

PROCEDURE ReadNum (cid: ChanId): CARDINAL;

    (* One decimal number per line.   *)

    CONST Digits = CharSet {'0'..'9'};

    VAR N: CARDINAL;
        ch: CHAR;

    BEGIN
        N := 0;
        REPEAT
            TextIO.ReadChar (cid, ch);
        UNTIL ch <> ' ';
        WHILE ch IN Digits DO
            N := 10*N + ORD(ch) - ORD('0');
            TextIO.ReadChar (cid, ch);
        END (*IF*);
        TextIO.SkipLine (cid);
        RETURN N;
    END ReadNum;

(****************************************************************)

PROCEDURE ReadData;

    (* Data arrives column by column, with a redundant column   *)
    (* heading that needs to be discarded.                      *)

    VAR row, col: CARDINAL;
        cid: ChanId;  res: SeqFile.OpenResults;

    BEGIN
        SeqFile.OpenRead (cid, InFileName, SeqFile.read, res);
        FOR col := 0 TO 15 DO
            FOR row := 0 TO 3 DO
                Data[row, col] := ReadNum (cid);
            END (*FOR*);
        END (*FOR*);
        SeqFile.Close (cid);
    END ReadData;

(****************************************************************)
(*                           OUTPUT                             *)
(****************************************************************)

PROCEDURE WriteHex1 (cid: ChanId;  N: CARDINAL);

    BEGIN
        IF N < 10 THEN
            TextIO.WriteChar (cid, CHR(N + ORD('0')));
        ELSE
            TextIO.WriteChar (cid, CHR(N - 10 + ORD('A')));
        END (*IF*);
    END WriteHex1;

(****************************************************************)

PROCEDURE WriteHex2 (cid: ChanId;  N: CARDINAL);

    BEGIN
        TextIO.WriteChar (cid, '0');
        WriteHex1 (cid, N DIV 16);
        WriteHex1 (cid, N MOD 16);
        TextIO.WriteChar (cid, 'H');
    END WriteHex2;

(****************************************************************)

PROCEDURE WriteCard (cid: ChanId;  N: CARDINAL);

    (* Write decimal. *)

    BEGIN
        IF N > 9 THEN
            WriteCard (cid, N DIV 10);
            N := N MOD 10;
        END (*IF*);
        TextIO.WriteChar (cid, CHR(ORD('0')+N));
    END WriteCard;

(****************************************************************)

PROCEDURE WriteData;

    VAR row, col: CARDINAL;
        cid: ChanId;  res: SeqFile.OpenResults;

    BEGIN
        SeqFile.OpenWrite (cid, OutFileName, SeqFile.write, res);
        FOR row := 0 TO 3 DO
            TextIO.WriteChar (cid, "{");
            FOR col := 0 TO 15 DO
                WriteCard (cid, Data[row, col]);
                IF col < 15 THEN
                    TextIO.WriteString (cid, ", ");
                END (*IF*);
            END (*FOR*);
            TextIO.WriteString (cid, "},");
            TextIO.WriteLn(cid);
        END (*FOR*);
        SeqFile.Close (cid);
    END WriteData;

(****************************************************************)

BEGIN
    ReadData;
    WriteData;
END MakeTable4x16.

