MODULE MakeTable8x6;

    (********************************************************)
    (* Reformats some data from NIST.FIPS.46-3 to make it   *)
    (* more useable.                                        *)
    (********************************************************)


IMPORT TextIO, SeqFile;

FROM IOChan IMPORT ChanId;

(****************************************************************)

CONST
    InFileName = "Indata.txt";
    OutFileName = "outdata.txt";

TYPE
    CharSet = SET OF CHAR;

VAR Data: ARRAY [0..7],[0..5] OF CARDINAL;

(****************************************************************)
(*                           INPUT                              *)
(****************************************************************)

PROCEDURE ReadNum (cid: ChanId): CARDINAL;

    (* One decimal number per line.   *)

    CONST Digits = CharSet {'0'..'9'};

    VAR N: CARDINAL;
        ch: CHAR;

    BEGIN
        N := 0;
        REPEAT
            TextIO.ReadChar (cid, ch);
        UNTIL ch <> ' ';
        WHILE ch IN Digits DO
            N := 10*N + ORD(ch) - ORD('0');
            TextIO.ReadChar (cid, ch);
        END (*IF*);
        TextIO.SkipLine (cid);
        RETURN N;
    END ReadNum;

(****************************************************************)

PROCEDURE ReadData;

    (* Data arrives column by column.                      *)

    VAR row, col: CARDINAL;
        cid: ChanId;  res: SeqFile.OpenResults;

    BEGIN
        SeqFile.OpenRead (cid, InFileName, SeqFile.read, res);
        FOR col := 0 TO 5 DO
            FOR row := 0 TO 7 DO
                Data[row, col] := ReadNum (cid);
            END (*FOR*);
        END (*FOR*);
        SeqFile.Close (cid);
    END ReadData;

(****************************************************************)
(*                           OUTPUT                             *)
(****************************************************************)

PROCEDURE WriteCard (cid: ChanId;  N: CARDINAL);

    (* Writes N in decimal. *)

    BEGIN
        IF N > 9 THEN
            WriteCard (cid, N DIV 10);
            N := N MOD 10;
        END (*IF*);
        TextIO.WriteChar (cid, CHR(N + ORD('0')));
    END WriteCard;

(****************************************************************)

PROCEDURE WriteData;

    VAR row, col: CARDINAL;
        cid: ChanId;  res: SeqFile.OpenResults;

    BEGIN
        SeqFile.OpenWrite (cid, OutFileName, SeqFile.write, res);
        FOR row := 0 TO 7 DO
            TextIO.WriteChar (cid, "{");
            FOR col := 0 TO 5 DO
                WriteCard (cid, Data[row, col]);
                IF col < 5 THEN
                    TextIO.WriteString (cid, ", ");
                END (*IF*);
            END (*FOR*);
            TextIO.WriteString (cid, "},");
            TextIO.WriteLn(cid);
        END (*FOR*);
        SeqFile.Close (cid);
    END WriteData;

(****************************************************************)

BEGIN
    ReadData;
    WriteData;
END MakeTable8x6.

